""" Paragon Python Module

Python interface to Calnex Paragon
This file supports Paragon-X, Paragon-100G and Paragon-neo.
Not all functions are available for all instruments.
See the Paragon Remote Control Manual for details.

Some functionality for Paragon-100G and Paragon-neo is implemented directly
in this wrapper and not in the instrument (for example any command that
involves exporting a file to a folder on a local PC). In this case, the
additional functionality is implemented in calnex_translation.py which
is expected to exist in the same folder as paragon.py.
calnex_translation.py makes use of calnexRestModified.py to communicate with
Paragon-100G / Paragon-neo. calnexRestModified.py also needs to exist in the
same folder as paragon.py
Note: calnexRestModified.py has been updated to support the translation
functionality. These updates may be integrated into the
Paragon-100G / Paragon-neo at some point.
"""
##############################################################################
# Copyright (c) Calnex Solutions Ltd 2008 - 2019
# The contents of this file should not be copied or distributed
# without permission being granted by Calnex Solutions Ltd.
#
# All rights reserved.
##############################################################################
# TODO:
#   Create a separate error handlng class
#   Add support for exportdata
##############################################################################
# Version History:
#   27 Nov 2019, GM: Added logging
#   02 Dec 2019, GM: Modified to support Python 2.7 (as well as 3)
#   11 Sep 2020, GM: Added translation for 802.1AS-REV profile and PFV rules.
#       Versions later than 10.41.07 have renamed -REV to 802.1AS-2020
#   11 Jan 2021, GM: Fixed a bug in the parsing of IDN information on connect

from __future__ import print_function

# Removed - this is Python 3 only.
# This was being used to avoid try-except pass cases where any exceptions
# are ignored. This made the code tidier but is not required.
# from contextlib import suppress

import socket
import sys
import shlex
import time
import os
import logging
import logging.handlers as handlers

# Private Members and Functions
# -----------------------------
_REV_STR = "$Revision: 27483 $"

# Expected reply constants
_PARAGON_READY = "__PARAGON__"
_PARAGON_ERROR = "ERROR: "

# Private module variables
# ------------------------
_last_err = ""
_paragon = None

# Debug enable/disable
_DEBUG = False


class ParagonSocket():
    """ Class to deal with the socket connection to the GUI """
    _socket = None

    def __init__(self, gui_address, gui_port):
        """ Create the socket and connect """
        self._socket = socket.socket()
        self._socket.connect((gui_address, gui_port))

    def receive(self):
        """ Read a line from the socket """
        global _last_err

        response = ""
        _last_err = ""

        # Create file from socket for convenience
        # and read each line from it
        file = self._socket.makefile()
        for line in file:
            # _debug("Received: %s" % line)
            # Errors are returned as a string beginning 'ERROR: '
            # Ready is signalled by '__PARAGON__' prompt
            # Anything else is considered a response
            if line.find(_PARAGON_ERROR) >= 0:
                # Got an error
                _last_err = line[len(_PARAGON_ERROR):len(line)]
            elif line.find(_PARAGON_READY) >= 0:
                # Got the ready prompt, OK to leave
                break
            else:
                # Got a response. Store it and continue
                response = response + line
        file.close()

        return response

    if sys.version_info < (3, 0):  # py2
        def send(self, data):
            """ Send data to the socket """

            # _debug("Sending %s... " % data)
            self._socket.sendall("%s\n" % data)
            # _debug("Done.")
            return self.receive().strip()
    else:  # py3
        def send(self, data):
            """ Send data to the socket """

            # _debug("Sending %s... " % data)
            self._socket.sendall(bytes("%s\n" % data, "UTF-8"))
            # _debug("Done.")
            return self.receive().strip()

    def close(self):
        """ Close the socket """
        self._socket.close()


class Paragon():
    """ Implements the Paragon API """

    _instrument_type = None
    _instrument_ip = ""
    _gui_ip = ""
    _instrument_port = 9990
    _gui_port = 9000
    _calnex_translator = None

    _translation_files_exist = False
    _socket = None
    logger = None

    _rev_str = _REV_STR

    def __init__(self, instrument_ip, gui_ip,
                 instrument_port=9990, gui_port=9000):
        """ Initialise an instance of the Paragon class
        Parameters:
            instrument_ip: str
                Paragon-X - The IP address of the Paragon-X
                Paragon-Neo, Paragon-100G - must be localhost
            gui_ip: str
                Paragon-X - The IP address of the Paragon-X GUI.
                            "localhost" if the GUI is running locally
                Paragon-Neo, Paragon-100G - The IP address of the instrument
            instrument_port: int
                Paragon-X - defaults to 9990. Should not need to be changed
                Paragon-Neo, Paragon-100G - NA
            gui_port: int
                Paragon-X - defaults to 9000. Only needs to be changed when
                            the GUI has been started with a different port
                Paragon-Neo, Paragon-100G - must be 9000
        """
        self._translation_files_exist = self._do_translation_files_exist()
        self._instrument_ip = instrument_ip
        self._gui_ip = gui_ip
        self._instrument_port = instrument_port
        self._gui_port = gui_port
        self._configure_logging()

    def _configure_logging(self):
        self.logger = logging.getLogger("paragon")
        self.logger.setLevel(logging.INFO)
        if not len(self.logger.handlers):
            ch = logging.StreamHandler()
            # ch.setLevel(logging.INFO)
            formatter = logging.Formatter(
                '%(name)20s %(levelname)-8s %(funcName)30s %(message)s')
            ch.setFormatter(formatter)
            self.logger.addHandler(ch)

            script_path = os.path.dirname(os.path.realpath(__file__))
            log_path = script_path

            fh = handlers.RotatingFileHandler(
                log_path+'/WrapperLogging.txt', 'a', 100000, 5, None, True)
            # fh.setLevel(logging.INFO)
            formatter = logging.Formatter(
                "%(asctime)s %(name)20s %(levelname)-8s %(funcName)30s %(message)s")
            fh.setFormatter(formatter)
            self.logger.addHandler(fh)

        self.logger.disabled = True

    def _do_translation_files_exist(self):
        """ Check that files required for translation exists
        Returns
            bool: True if the required files exist
        """
        files_ok = False
        wrapper_folder = os.path.dirname(os.path.realpath(__file__))
        if os.path.exists(os.path.join(wrapper_folder,
                                       "calnex_translation.py")):
            files_ok = True

        return files_ok

    def wrapper_logging(self, enable=True, level=20):
        """ Control command logging
        Parameters:
            enable: TRUE to enable logging; FALSE to disable it.
            level: The logging level above which a logging entry will be
                    generated.
                    The levels are debug=10, info=20, warning=30, error=40,
                    critical=50. So to see all info and above, set to 20
        Description:
            When logging is enabled, all commands are written to the log file
            and console.
            The log file is stored in the same folder in which the calling
            script is being run.
            The log file is stored in a text file called
                "wrapperLogging.txt".
        Usage:
            Call wrapper_logging(TRUE) at the point in our script where you
            want to log commands.
            Call wrapper_logging(FALSE) when you no longer want to log.
        Note: The log file will only be closed when either the calling script
            terminates.
        """
        if enable:
            self.logger.disabled = False
            self.logger.setLevel(level)
            self.logger.info(
                "-------------------- Logging Enabled --------------------")
        else:
            self.logger.info(
                "-------------------- Logging Disabled -------------------")
            self.logger.disabled = True

    def connect(self, demo_mode_enable=0):
        """ Makes a connection to the specified GUI and instrument.
        Parameters:
            demo_mode_enable: int
                Enable demo mode
        Raises:
            ConnectionRefusedError
                If it is not possible to connect to the Paragon instrument
                because the isntrument rejected the connection or the
                instrument / GUI is not visible on the network or running
            Exception
                If it is not possible to connect to the instrument
                for an unknown reason
        """
        # Modified to follow the same process as the Tcl wrapper.
        # Also updated to be more robust especially around
        # connecting in demo mode

        global _last_err

        status = 0
        error = ""

        # _gui_connect(gui_address, guiport)
        self._gui_connect()

        try:
            # _remote_control_on(rc_enable)
            self._remote_control_on(1, demo_mode_enable)
            # self._remote_control_on(1)

            # conn_status = _is_connected("")
            # conn_status = self._is_instrument_connected("")
            conn_status = self.is_connected("")
            # print("connect: conn_status = %d" % conn_status)
            if conn_status < 2:
                # inst_status = _is_connected(inst_address)
                inst_status = self._is_instrument_connected()
                # print ("connect: inst_status = %d" % inst_status)
                if conn_status == 1 and inst_status == 0:
                    curr_ip = self.paragonget("RmtIpAddress")
                    err_str = \
                        "ERROR: another instrument (%s) is already connected"
                    error = err_str % curr_ip
                elif inst_status == 0:
                    # _instrument_connect(inst_address, instport)
                    self._instrument_connect()
                    # _connected = True
            # else:
                # _connected = True

        except Exception:
            status = 1

        if error != "":
            # raise Exception("connect: " + error)
            _last_err = error
            self.logger.error(error)
            _check_for_error("connect")

        if status == 0:
            # if _connected == 1:
            # _debug("connected")
            self.logger.info("Connected")
            # idn_string = self.paragonget("Idn")
            idn_string = self._get_instrument_details()
            rev_parts = self._rev_str.split()
            # print("Connected to " + _instrument_type)
            print(idn_string + ' (' + rev_parts[1] + ')')
            self._connect_translator()
        else:
            # _debug("not connected")
            self.logger.info("Not connected")
            err = _last_err
            self.disconnect()
            _last_err = err
            _check_for_error("connect")

    def _get_instrument_details(self):
        """
        Figure out the type of instrument to which we are connected.
        Return the full Idn string
        """
        self.logger.info("")
        idn_string = self.paragonget("Idn")
        idn_sections = idn_string.split(';')
        idn_parts = idn_sections[0].split(',')
        self._instrument_type = idn_parts[1]
        self._instrument_serial_number = idn_parts[2]
        server_parts = idn_parts[3].split("=")
        x, x_num, major, minor = server_parts[1].split(".")
        self._version_major = int(major)
        self._version_minor = int(minor)

        return idn_string

    def _connect_translator(self):
        """ Initialise the translator """
        self.logger.debug("")
        # self._instrument_type = self._get_instrument_type()
        if self._is_translator_available():
            import calnex_translation as ct
            self._calnex_translator = \
                ct.CalnexTranslator(self._instrument_type, self._instrument_ip,
                                    self._gui_ip)

    def _is_translator_available(self):
        """ Returns true if the translator files are present """
        # self.logger.debug("")
        return self._translation_files_exist

    # Get the instument type from the Idn string
    def _get_instrument_type(self):
        """ Return the type of instrument to which we are connected """
        self.logger.info("")
        idn_string = self.paragonget("Idn")
        # print(idnString)
        idn_parts = idn_string.split(',')
        return idn_parts[1]

    def _gui_connect(self):
        """ Connect to the GUI """
        # global _socket
        print("Connecting to %s:%s... " % (self._gui_ip, self._gui_port))
        self.logger.info(
            "Connecting to %s:%s... ", self._gui_ip, self._gui_port)

        # Create a new socket and connect to the GUI
        self._socket = ParagonSocket(self._gui_ip, self._gui_port)

        # print("Success: connection to the GUI established")

        # Flush and check for errors
        self._socket.receive()
        _check_for_error("gui_connect")

    def _is_instrument_connected(self, ip_addr="UNDEFINED"):
        """ Returns 1 if the instrument is connected
        Parameters:
            ip_addr: str
                The IP address of the isntrument
        Returns:
            int: 1 if instrument is connected
        """
        self.logger.debug(ip_addr)
        if ip_addr == "UNDEFINED":
            ip_addr = self._instrument_ip
        response = self._socket.send(
            "paragon.remoteprint(paragon.isconnected(\"%s\"))" % ip_addr)
        _check_for_error("is_connected")
        self.logger.debug("----Response: %s", response)
        return int(response)

    def _is_gui_connected(self):
        """ Returns 1 if the GUI is connected """
        self.logger.debug("")

        if self._socket is None:
            resp = 0
            # print ("isguiconnected: No socket")
        else:
            # We still have a socket open but the connection may have been
            # killed by the user.
            # Try to figure out if the connection is still active
            resp = 1
            try:
                # print ("isguiconnected: Sending remote_control_on")
                self._remote_control_on(1)
            except Exception:
                resp = 0

        # print ("isguiconnected: resp = %s" % resp)
        self.logger.debug("----Response: %s", resp)
        return resp

    def _instrument_connect(self):
        """ Connect to the isntrument """
        print("Connecting to instrument %s:%s... " % (self._instrument_ip,
                                                      self._instrument_port))
        self.logger.debug("Connecting to instrument %s:%s... ",
                          self._instrument_ip, self._instrument_port)
        self._socket.send(
            "paragon.connect(\"%s\",\"%s\")" % (self._instrument_ip,
                                                self._instrument_port))
        _check_for_error(
            "_instrument_connect %s:%s" % (self._instrument_ip,
                                           self._instrument_port))

    def _instrument_disconnect(self):
        """ Dicsonnect from the instrument """
        self.logger.debug("")
        self._socket.send("paragon.disconnect()")
        _check_for_error("_instrument_disconnect")

    def _query_errors(self):
        """ Check instrument for errors """
        self.logger.debug("")
        self._socket.send("paragon._queryerrors()")
        _check_for_error("_query_errors")

    def _remote_control_on(self, rc_enable, demo_mode_enable=0):
        """ Enable remote control """
        # TODO: Change coming in 27X patch release
        # self._socket.send(
        # "paragon.on(%s,%s)" % (rc_enable, demo_mode_enable))
        self.logger.debug("")
        self._socket.send("paragon.on(%s)" % rc_enable)
        _check_for_error("_remote_control_on")

    def _remote_control_off(self):
        """ Disable remote control """
        self.logger.debug("")
        self._socket.send("paragon.off()")
        _check_for_error("_remote_control_off")

    def _set_operating_mode(self, operating_mode):
        """ Set the OperatingMode
        Parameters:
            operating_mode: str
                See the Remote Control Manual for details
        """
        self.logger.debug(operating_mode)
        self._socket.send(
            "paragon.CSettingParameter(\"__SET____#__OperatingMode__#__%s\")" %
            operating_mode)
        _check_for_error("_set_operating_mode %s" % operating_mode)

    def _close_connection(self):
        """ Close the socket connection to the GUI """
        self.logger.debug("")
        self._socket.close()
        self._socket = None

    def disconnect(self):
        """ Disconnect """
        self.logger.info("")
        try:
            self._instrument_disconnect()
        except Exception:
            pass
        try:
            self._remote_control_off()
        except Exception:
            pass
        try:
            self._close_connection()
        except Exception:
            pass

    def is_connected(self, instrument_ip=""):
        """ Determines if there is a connection to the GUI and isntrument
        Parameters:
            instrument_ip: str
                The IP address of the instrument to be checked
        Returns:
            1 if there is a connection to the GUI and instrument
            0 otherwise
        """
        self.logger.info("")
        resp = 0

        # Send generates an error if not connected to the GUI.
        # For isconnected, we should simply return false
        if instrument_ip == "":
            ip = self._instrument_ip
        else:
            ip = instrument_ip

        if ip == "":
            resp = 0
        else:
            gui_conn = self._is_gui_connected()
            # print ("isconnected: gui_conn = %d" % gui_conn)
            if gui_conn == 0:
                resp = 0
            else:
                resp = self._is_instrument_connected(ip)
                _check_for_error("isconnected")

        self.logger.info("---Response: %s", str(resp))
        return resp

    def paragonset(self, arg_str, value=None):
        """ Set an instrument setting.
        Most of the instrument's settings are controlled
        using by this command.
        (Refer to the remote control manual for the complete list)
        Parameters:
            arg_str: str
                Setting to set (e.g "Physical LineRate").
            value: str
                The value to set.
        Raises:
            Exception
        """
        if value:
            command = arg_str + " " + value
        else:
            command = arg_str

        self.logger.info(command)
        # _debug("paragonset " + command)
        tokens = command.split(' ')
        is_bypass = False
        wire_cmd = "paragon.CSettingParameter(\"__SET__"

        # Parse the incoming command such that specific actions can be taken
        # for specific commands.
        # In most cases, when a command is found,
        # it will bypass further processing
        if self._instrument_type == "Paragon-X":
            is_bypass = _rst_translation(arg_str)
        if self._instrument_type == "Paragon-X" and (self._version_major > 41 or (self._version_major == 41 and self._version_minor > 7)):
            is_bypass = _standards_profile_translation(arg_str, value)
        if self._is_translator_available():
            if tokens[0] == "Cat":
                # _debug("Translating...")
                self.logger.debug("---Translating Cat command...")
                is_bypass = self._calnex_translator.cat_set(command)

        # Only send the command is it hasn't already been dealt with
        if is_bypass is False:
            wire_cmd += _build_cmd_from_args(*shlex.split(arg_str))
            if value is not None:
                wire_cmd += "__#__" + value
            wire_cmd += "\")"

            # Some paragonset Cat commands may return an error
            # while in live mode. This is caused by the rolling re-calculation.
            # So, make two further attempts to retry if this happens
            # This has been seen with Cat SelectSlot and Cat <metric> Enable
            # The errors are "Undefined command" and "Invalid parameter"
            retries = 10
            if tokens[0] == "Cat":
                retries = 100
            for retry in range(retries):
                resp = self._socket.send(wire_cmd)
                # Look for "Metric is not calculated"; if present, retry
                # If not, call chkForError
                # If at the end of retries, call chkForError
                found = False
                if resp.startswith("Undefined command") or \
                        resp.startswith("Invalid parameter"):
                    found = True
                if not found or retry == retries:
                    _check_for_error("paragonset %s" % wire_cmd)
                    break

                self.logger.debug("RETRY---> paragonset %s --- retry %s: %s",
                                  command, str(retry), resp)
                time.sleep(1)

    def paragonget(self, arg_str):
        """ Query an instrument setting.
        Most of the instrument's settings are read
        using by this command.
        (Refer to the remote control manual for the complete list)
        Paramters:
            arg_str: str
                Setting to query (e.g "Physical LineRate").
        Returns:
            str: The setting value.
        Raises:
            Exception
        """
        # _debug("paragonget " + arg_str)
        self.logger.info(arg_str)
        tokens = arg_str.split(' ')
        is_bypass = False
        wire_cmd = "paragon.CSettingParameter(\"__GET__"
        wire_cmd += _build_cmd_from_args(*shlex.split(arg_str)) + "\")"

        if self._is_translator_available():
            if tokens[0] == "Cat":
                # _debug("Translating...")
                self.logger.debug("---Translating...")
                resp = self._calnex_translator.cat_get(arg_str)
                is_bypass = resp["bypass"]
                response = resp["response"]
        else:
            # _debug("NOT translating...")
            self.logger.debug("---Not translating...")

        # Only send the command if it hasn't already been dealt with
        if is_bypass is False:
            # paragonget Cat <metric> <stat> may return
            # "Metric is not calculated"
            # while in live mode. This is caused by the rolling re-calculation.
            # So, retry if this happens
            retries = 100
            for retry in range(retries):
                self._socket.send(wire_cmd)
                _check_for_error("paragonget %s" % wire_cmd)

                response = self._socket.send(
                    "print(paragon.CSettingParameter_resp())")
                # Look for "Metric is not calculated"; if present, retry
                # If not, call chkForError
                # If at the end of retries, call chkForError
                found = False
                if ("Metric is not calculated" in response or
                        "Type of metric doesn't exist" in response):
                    found = True

                if not found or retry == retries-1:
                    _check_for_error("paragonget %s" % wire_cmd)
                    # response = self._socket.send(
                    #     "print(paragon.CSettingParameter_resp())")
                    # _check_for_error("paragonget %s" % response)
                    break
                self.logger.debug(
                    "RETRY-> paragonget %s - retry %s: %s",
                    arg_str, str(retry), response)
                time.sleep(1)

            # Ignore the type ID character at the start of the response
            response = response[1:len(response)]

        self.logger.info("---Response: %s", response)
        return response

    def start_packet_capture(self):
        """ Starts an ALL byte capture """
        self.logger.info("")
        self._socket.send("paragon.startpacketcapture()")
        _check_for_error("startpacketcapture")

    def start_timing_capture(self):
        """ Starts a timing capture """
        self.logger.info("")
        self._socket.send("paragon.starttimingcapture()")
        _check_for_error("starttimingcapture")

    def stop_capture(self):
        """ Stops the current capture """
        self.logger.info("")
        self._socket.send("paragon.stopcapture()")
        _check_for_error("stopcapture")

    def start_impairment(self):
        """ Starts the impairment as configured in the settings. """
        self.logger.info("")
        self._socket.send("paragon.startimpairment()")
        _check_for_error("startimpairment")

    def stop_impairment(self):
        """ Stops the currently running impairment """
        self.logger.info("")
        self._socket.send("paragon.stopimpairment()")
        _check_for_error("stopimpairment")

    def start_tod_capture(self):
        """ Starts a time of day capture """
        self.logger.info("")
        self._socket.send("paragon.starttodcapture()")
        _check_for_error("starttodcapture")

    def stop_tod_capture(self):
        """ Stops a time of day capture """
        self.logger.info("")
        self._socket.send("paragon.stoptodcapture()")
        _check_for_error("stoptodcapture")

    def set_sonet_owsts1_trib(self, trib):
        """ Select the SONET STS1 tributary to impair.
        Parameters
            trib: str
                STS1 tributary number.
        Raises
            Exception
        """
        self.logger.info(trib)
        cmd_str = \
            "paragon.CSettingParameter(\"__SET____#___SonetSts1Trib__#__%s\")"
        self._socket.send(cmd_str % trib)
        _check_for_error("setsonetowsts1trib %s" % trib)

    def set_sonet_owvt_trib(self, trib):
        """ Select the SONET VT1.5 tributary to impair.
        Parameters
            trib: str
                VT1.5 tributary number.
        Raises
            Exception
        """
        self.logger.info(trib)
        cmd_str = \
            "paragon.CSettingParameter(\"__SET____#___SonetVtTrib__#__%s\")"
        self._socket.send(cmd_str % trib)
        _check_for_error("setsonetowvttrib %s" % trib)

    def set_sonet_delay(self, delay):
        """ Set the SONET delay """
        self.logger.info(delay)
        cmd_str = \
            "paragon.CSettingParameter(\"__SET____#___SonetDelay__#__%s\")"
        self._socket.send(cmd_str % delay)
        _check_for_error("setsonetdelay %s" % delay)

    def export_data(self, filename):
        """ Saves the displayed captured data to a file.
        Parameters:
            filename: str
                Path and filename for a file on the local PC's
                file system. filename must be suffixed with a
                valid filename extension - ".csv" or ".cpd"
        Raises:
            Exception
        """
        self.logger.info(filename)
        global _last_err

        if (self._instrument_type == "Paragon-Neo" or
                self._instrument_type == "Paragon-100G"):
            _last_err = "Only supported for Paragon-X"
        else:
            filename = filename.replace("\\", "/")
            self._socket.send("paragon.exportdata(\"%s\")" % filename)

        _check_for_error("exportdata %s" % filename)

    def import_data(self, filename):
        """ Loads the captured data from a file.
        This is the equivalent to using the GUI toolbar menu item File>Import.
        This is not port specific and in the bi-directional context is not
        intended to be used for loading impairment data for replay.
        Parameters:
            filename: str
                Path and filename for a file on the local PC's
                file system. filename must be suffixed with a
                valid filename extension - ".csv" or ".cpd"
        Raises:
            Exception
        """
        self.logger.info(filename)
        global _last_err

        if (self._instrument_type == "Paragon-Neo" or
                self._instrument_type == "Paragon-100G"):
            _last_err = "Only supported for Paragon-X"
        else:
            filename = filename.replace("\\", "/")
            self._socket.send("paragon.importdata(\"%s\")" % filename)

        _check_for_error("importdata %s" % filename)

    def store(self, filename):
        """ Stores the instrument settings to the file passed in using the
        filename parameter. Settings are stored in the file system of the
        machine hosting the GUI.
        The stored settings file can be recalled using the 'recall' command.
        If the specified file already exists, then it will be overwritten.
        Parameters:
            filename: str
                Path and filename for a file on the local PC's
                file system.
        Raises:
            Exception
        """
        self.logger.info(filename)
        global _last_err

        if (self._instrument_type == "Paragon-Neo" or
                self._instrument_type == "Paragon-100G"):
            _last_err = "Only supported for Paragon-X"
        else:
            filename = filename.replace("\\", "/")
            self._socket.send("saveSettings(\"%s\")" % filename)

        _check_for_error("store %s" % filename)

    def recall(self, filename):
        """ Recalls the instrument settings from the file specified by the
        filename parameter. Settings are stored in the file system of the
        machine hosting the GUI.
        A stored settings file can be created by from the Paragon GUI
        using Save on the Setup menu.
        Parameters:
            filename: str
                Path and filename for a file on the local PC's file system.
        Raises:
            Exception
        """
        self.logger.info(filename)
        global _last_err

        if (self._instrument_type == "Paragon-Neo" or
                self._instrument_type == "Paragon-100G"):
            _last_err = "Only supported for Paragon-X"
        else:
            filename = filename.replace("\\", "/")
            self._socket.send("loadSettings(\"%s\")" % filename)

        _check_for_error("recall %s" % filename)

    def import_impairment_data(self, port, filename):
        """ Loads the capture data for replay.
        It will load files specifically for replaying delay profiles against
        defined message types on defined ports.
        The loaded file will be replayed against incoming traffic on the
        specified port.
        This command is the equivalent of "Import-Port1" and
        "Import-Port2" buttons in the Add Impairments and Delay window
        Paramters:
            port: str
                The Ethernet port (1 or 2).
            filename: str
                Path to a capture file on the local PC's file system.
        Raises:
            Exception
        """
        self.logger.info("%s %s", port, filename)

        global _last_err

        if (self._instrument_type == "Paragon-Neo" or
                self._instrument_type == "Paragon-100G"):
            _last_err = "Only supported for Paragon-X"
        else:
            filename = filename.replace("\\", "/")
            self._socket.send(
                "paragon.importimpairmentdata(\"%s\",\"%s\")"
                % (port, filename))

        _check_for_error(
            "importimpairmentdata %s %s" % (port, filename))

    def wait_for_cat(self):
        """ Wait for the completion of the current CAT operation.
        If a script is waiting for a previous open operation or waiting for end
        of a previous calculation then this command should be always used.
        See the remote control manual for guidelines on when
        this should be used.
        """
        self.logger.info("")
        instrument_type = self._instrument_type

        if (instrument_type == "Paragon-100G" or
                instrument_type == "Paragon-Neo"):
            if self._is_translator_available():
                self.logger.debug("----Translating...")
                self._calnex_translator.cat_wait()
            else:
                status = self.paragonget("Cat Status")[0]
                while status == "1":
                    status = self.paragonget("Cat Status")[0]
                    time.sleep(1)
        else:
            status = self.paragonget("Cat Status")[0]
            while status == "1":
                status = self.paragonget("Cat Status")[0]
                time.sleep(1)

        # print("Wait operation completed")
        self.logger.info("----Wait operation complete")


# def _debug(msg):
#    if _DEBUG is True:
#        print(msg)
#        sys.stdout.flush()


def _check_for_error(label):
    if _last_err != "":
        raise Exception("%s : %s" % (label, _last_err))


def _build_cmd_from_args(*args):
    cmd = ""
    for arg in args:
        # Indexed: 1st char is '#'
        if arg[0] == '#':
            arg = "__IDX____#__" + arg[1:len(arg)]
        # Append argument
        cmd = cmd + "__#__" + arg
    return cmd


###############################################################################
# The following routines are used to make the Rst process more robust
# There are occasions where Rst has issues when the instrument is active
# These routines try to make sure that the box is in an idle state before the
# Rst is sent.
# Note: The Rst command is intercepted in paragonset; anything that is active
# on the instrument is then stopped and then the Rst is passed on.
# Since this file is also used for P100G, a number of the commands may fail
# if any given feature has not yet been implemented. This is avoided by making
# sure that errors are ignored in the appropriate places


def _stop_wander():
    """ Check to see if any wander generation is running
        If so, stop it
    """
    # _debug("stop_wander")
    _paragon.logger.debug("")
    is_freq_offset_running = "FALSE"
    tol_single_state = "IDLE"
    tol_table_state = "IDLE"
    tol_mtie_tdev_state = "IDLE"
    synce_transient_state = "IDLE"
    transfer_single_state = "IDLE"
    transfer_table_state = "IDLE"
    transfer_tdev_state = "IDLE"

    try:
        is_freq_offset_running = paragonget(
            "WanderGeneration FrequencyOffset Enable")
    except Exception:
        pass
    try:
        tol_single_state = paragonget(
            "WanderGeneration Tolerance Single State")
    except Exception:
        pass
    try:
        tol_table_state = paragonget("WanderGeneration Tolerance Table State")
    except Exception:
        pass
    try:
        tol_mtie_tdev_state = paragonget(
            "WanderGeneration Tolerance MtieTdev State")
    except Exception:
        pass
    try:
        synce_transient_state = paragonget(
            "WanderGeneration SyncETransient State")
    except Exception:
        pass
    try:
        transfer_single_state = paragonget(
            "WanderGeneration Transfer Single State")
    except Exception:
        pass
    try:
        transfer_table_state = paragonget(
            "WanderGeneration Transfer Table State")
    except Exception:
        pass
    try:
        transfer_tdev_state = paragonget(
            "WanderGeneration Transfer TDEV State")
    except Exception:
        pass

    if is_freq_offset_running != "FALSE":
        paragonset("WanderGeneration FrequencyOffset Enable", "FALSE")

    if tol_single_state != "IDLE":
        paragonset("WanderGeneration Tolerance Single Enable", "FORCESTOP")

    if tol_table_state != "IDLE":
        paragonset("WanderGeneration Tolerance Table Enable", "FORCESTOP")

    if tol_mtie_tdev_state != "IDLE":
        paragonset("WanderGeneration Tolerance MtieTdev Enable", "STOP")

    if synce_transient_state != "IDLE":
        paragonset("WanderGeneration SyncETransient Enable", "STOP")

    if transfer_single_state != "IDLE":
        try:
            paragonset(
                "WanderGeneration Transfer Single CalibrateEnable",
                "FORCESTOP")
        except Exception:
            pass
        try:
            paragonset(
                "WanderGeneration Transfer Single GenerateEnable",
                "FORCESTOP")
        except Exception:
            pass

    if transfer_table_state != "IDLE":
        try:
            paragonset(
                "WanderGeneration Transfer Table CalibrateEnable",
                "FORCESTOP")
        except Exception:
            pass
        try:
            paragonset(
                "WanderGeneration Transfer Table GenerateEnable", "FORCESTOP")
        except Exception:
            pass

    if transfer_tdev_state != "IDLE":
        paragonset("WanderGeneration Transfer TDEV Enable", "STOP")


def _stop_packet_generation():
    """ Check to see if any packet generation is running
        If so, stop it
    """
    # _debug("stopPacketGeneration")
    _paragon.logger.debug("")
    is_port0_running = "FALSE"
    is_port1_running = "FALSE"

    try:
        is_port0_running = paragonget("PacketGeneration #0 Enable")
    except Exception:
        pass
    try:
        is_port1_running = paragonget("PacketGeneration #1 Enable")
    except Exception:
        pass

    if is_port0_running != "FALSE":
        paragonset("PacketGeneration #0 Enable", "FALSE")
    if is_port1_running != "FALSE":
        paragonset("PacketGeneration #1 Enable", "FALSE")


def _stop_jitter():
    """ Check to see if any jitter generation is running
        If so, stop it
    """
    # _debug("stop_jitter")
    _paragon.logger.debug("")
    tol_single_state = "IDLE"
    tol_table_state = "IDLE"
    max_tol_table_state = "IDLE"

    try:
        tol_single_state = paragonget("Jitter Tolerance Single State")
    except Exception:
        pass
    try:
        tol_table_state = paragonget("Jitter Tolerance Table State")
    except Exception:
        pass
    try:
        max_tol_table_state = paragonget("Jitter MaxTolerable Table State")
    except Exception:
        pass

    if tol_single_state != "IDLE":
        paragonset("Jitter Tolerance Single Enable", "FORCESTOP")
    if tol_table_state != "IDLE":
        paragonset("Jitter Tolerance Table Enable", "FORCESTOP")
    if max_tol_table_state != "IDLE":
        paragonset("Jitter MaxTolerable Table Enable", "FORCESTOP")


def _stop_mse():
    """ Check to see if MSE is running
        If so, stop it
    """
    # _debug("stopMSE")
    _paragon.logger.debug("")
    is_master0_running = "FALSE"
    is_master1_running = "FALSE"
    is_slave_running = "FALSE"

    try:
        is_master0_running = paragonget("MasterSlave Master #0 Enabled")
    except Exception:
        pass
    try:
        is_master1_running = paragonget("MasterSlave Master #1 Enabled")
    except Exception:
        pass
    try:
        is_slave_running = paragonget("MasterSlave Slave Enabled")
    except Exception:
        pass

    if is_master0_running != "FALSE":
        paragonset("MasterSlave Master #0 Enabled", "FALSE")
    if is_master1_running != "FALSE":
        # Looks like the get doesn't return the correct status so ignore errors
        try:
            paragonset("MasterSlave Master #1 Enabled", "FALSE")
        except Exception:
            pass

    if is_slave_running != "FALSE":
        try:
            paragonset("MasterSlave Slave Enabled", "FALSE")
        except Exception:
            pass


def _stop_capture():
    """ Stop any running capture """
    # _debug("stop_capture")
    _paragon.logger.debug("")
    stopcapture()


def _stop_impairment():
    """ Stop any running impairments """
    # _debug("stop_impairment")
    _paragon.logger.debug("")
    try:
        stopimpairment()
    except Exception:
        pass


def _do_reset_preparation():
    """ Before executing Rst, make sure that the instrument is idle
    - If a capture is in progress then we are not idle.
    - If Thru mode is active we are not idle.
    - If import-export is active we are not idle.
        We can't check for this and so we ignore it - unlikely to happen
    - If MSE is active we are not idle.
    - If impairments are running we are not idle.
    - If wander generation is active we are not idle.
    - If Jitter generation is active we are not idle.
    - If Packet Generation is active we are not idle.
    """

    # _debug("do_reset_preparation")
    _paragon.logger.debug("")

    # Stop anything that is running
    _stop_capture()
    _stop_impairment()
    _stop_wander()
    _stop_packet_generation()
    _stop_jitter()
    _stop_mse()


def _rst_translation(arguments):
    """ Handle Rst
    """
    _paragon.logger.debug(arguments)
    # We want bypass set to false.
    # Once we are done with the other reset actions,
    # we want to execute the Rst normally
    bypass = False

    if "Rst" in arguments:
        # _debug("rst_translation." + arguments + ".")
        _do_reset_preparation()

    return bypass

def _standards_profile_translation(command, value):
    """
    For Paragon-X versions after X.10.41.07, the 802.1AS-REV_(GPTP) profile
    has been renamed as 802.1AS-2020_(GPTP). To maintain backwards compatibility,
    replace the old name with the new name.
    Note: The PFV rules file has also changed name. Assume that using the Paragon-X
    version will cover the majority of users (as opposed to checking the PFV version)
    """
    if value:
        command_string = command + " " + value
    else:
        command_string = command
    _paragon.logger.debug(command_string)

    bypass = False
    command_string = command_string.upper()
    if "802.1AS-REV" in command_string:
        if "PFV" in command_string:
            _paragon.paragonset(command, "802.1AS-2020_(GPTP).xml")
        else:
            _paragon.paragonset(command, "802.1AS-2020_(GPTP)")

        bypass = True
    return bypass


# PUBLIC functions
# ================

def wrapper_logging(enable=True, level=20):
    """ Control command logging
    Parameters:
        enable: TRUE to enable logging; FALSE to disable it.
        level: The logging level above which a logging entry will be
                generated.
                The levels are debug=10, info=20, warning=30, error=40,
                critical=50. So to see all info and above, set to 20
    Description:
        When logging is enabled, all commands are written to the log file
        and console.
        The log file is stored in the same folder in which the calling
        script is being run.
        The log file is stored in a text file called
            "wrapperLogging.txt".
    Usage:
        Call wrapper_logging(TRUE) at the point in our script where you
        want to log commands.
        Call wrapper_logging(FALSE) when you no longer want to log.
    Note: The log file will only be closed when either the calling script
        terminates.
    """
    _paragon.wrapper_logging(enable, level)


def connect(inst_address, gui_address="localhost",
            guiport=9000, instport=9990, demo_mode_enable=0, rc_enable=1):
    """ Makes a connection to the specified GUI and instrument.
    Parameters
        instAddress: str, required
            The IP address of the Paragon instrument.
        guiAddress: str, optional
            The IP address of the PC that is hosting the GUI.
            If this parameter is not specified then a connection to the local
            machine is assumed.
        guiport: int, optional
            The remote control TCP Port used by the Paragon client application.
            If this parameter is not specified then port 9000 is assumed.
        instPort: int, optional
            The TCP Port on the Paragon instrument.
            If this parameter is not specified then port 9990 is assumed.
        demo_mode_enable: int
            (PX-5008) Add the ability to open the Paragon-X GUI remotely in
            demo mode
    Raises
        ConnectionRefusedError
            If it is not possible to connect to the Paragon instrument
            because the isntrument rejected the connection or the
            instrument / GUI is not visible on the network or running
        Exception
            If it is not possible to connect to the instrument
            for an unknown reason
    """
    # Modified to follow the same process as the Tcl wrapper.
    # Also updated to be more robust especially around connecting in demo mode
    global _paragon

    _paragon = Paragon(inst_address, gui_address, instport, guiport)
    _paragon.connect(demo_mode_enable)


def disconnect():
    """ Disconnects the currently connected instrument """

    _paragon.disconnect()
    print("Disconnected.")


def isconnected(instrument_ip=""):
    """ Returns 1 if there is a connection to the instrument; 0 otherwise """
    resp = _paragon.is_connected(instrument_ip)
    return resp


def paragonget(arg_str):
    """ Query an instrument setting.
    Most of the instrument's settings are read
    using by this command.
    (Refer to the remote control manual for the complete list)
    Paramters:
        arg_str: str
            Setting to query (e.g "Physical LineRate").
    Returns:
        str: The setting value.
    Raises:
        Exception
    """
    response = _paragon.paragonget(arg_str)
    return response


def paragonset(arg_str, value=None):
    """ Set an instrument setting.
    Most of the instrument's settings are controlled
    using by this command.
    (Refer to the remote control manual for the complete list)
    Parameters:
        arg_str: str
            Setting to set (e.g "Physical LineRate").
        value: str
            The value to set.
    Raises:
        Exception
    """
    _paragon.paragonset(arg_str, value)


def startpacketcapture():
    """ Starts an ALL byte capture """
    _paragon.start_packet_capture()


def starttimingcapture():
    """ Starts a timing capture """
    _paragon.start_timing_capture()


def stopcapture():
    """ Stops the current capture """
    _paragon.stop_capture()


def startimpairment():
    """ Starts the impairment as configured in the settings. """
    _paragon.start_impairment()


def stopimpairment():
    """ Stops the currently running impairment """
    _paragon.stop_impairment()


def starttodcapture():
    """ Starts a time of day capture """
    _paragon.start_tod_capture()


def stoptodcapture():
    """ Stops a time of day capture """
    _paragon.stop_tod_capture()


def setsonetowsts1trib(trib):
    """ Select the SONET STS1 tributary to impair.
    Parameters
        trib: str
            STS1 tributary number.
    Raises
        Exception
    """
    _paragon.set_sonet_owsts1_trib(trib)


def setsonetowvttrib(trib):
    """ Select the SONET VT1.5 tributary to impair.
    Parameters
        trib: str
            VT1.5 tributary number.
    Raises
        Exception
    """
    _paragon.set_sonet_owvt_trib(trib)


def setsonetdelay(delay):
    """ Set the SONET delay """
    _paragon.set_sonet_delay(delay)


def exportdata(filename):
    """ Saves the displayed captured data to a file.
    Parameters
        filename: str
            Path and filename for a file on the local PC's
            file system. filename must be suffixed with a
            valid filename extension - ".csv" or ".cpd"
    Raises
        Exception
    """
    _paragon.export_data(filename)


def importdata(filename):
    """ Loads the captured data from a file.
    This is the equivalent to using the GUI toolbar menu item File>Import.
    This is not port specific and in the bi-directional context is not intended
    to be used for loading impairment data for replay.
    Parameters
        filename: str
            Path and filename for a file on the local PC's
            file system. filename must be suffixed with a
            valid filename extension - ".csv" or ".cpd"
    Raises
        Exception
    """
    _paragon.import_data(filename)


def store(filename):
    """ Stores the instrument settings to the file passed in using the
    filename parameter. Settings are stored in the file system of the
    machine hosting the GUI.
    The stored settings file can be recalled using the 'recall' command.
    If the specified file already exists, then it will be overwritten.
    Parameters:
        filename: str
            Path and filename for a file on the local PC's
            file system.
    Raises:
        Exception
    """
    _paragon.store(filename)


def recall(filename):
    """ Recalls the instrument settings from the file specified by the filename
    parameter. Settings are stored in the file system of the machine hosting
    the GUI. A stored settings file can be created by from the Paragon GUI
    using Save on the Setup menu.
    Parameters:
        filename: str
            Path and filename for a file on the local PC's file system.
    Raises:
        Exception
    """
    _paragon.recall(filename)


def importimpairmentdata(port, filename):
    """ Loads the capture data for replay.
    It will load files specifically for replaying delay profiles against
    defined message types on defined ports.
    The loaded file will be replayed against incoming traffic on the
    specified port.
    This command is the equivalent of "Import-Port1" and
    "Import-Port2" buttons in the Add Impairments and Delay window
    Paramters:
        port: str
            The Ethernet port (1 or 2).
        filename: str
            Path to a capture file on the local PC's file system.
    Raises:
        Exception
    """
    _paragon.import_impairment_data(port, filename)


def waitforcat():
    """ Wait for the completion of the current CAT operation.
    If a script is waiting for a previous open operation or waiting for end
    of a previous calculation then this command should be always used.
    See the remote control manual for guidelines on when
    this should be used.
    """
    _paragon.wait_for_cat()
